from bs4 import BeautifulSoup
from typing import List, Optional
from datetime import datetime

from noire.models.moderation import ModerationRequest, ModerationDetails


def extract_moderation_requests(raw_html: str) -> List[ModerationRequest]:
    results = []
    soup = BeautifulSoup(raw_html, "html.parser")
    held_messages = soup.find_all("table", border="1")

    for message in held_messages:
        sender = message.find("strong", string="From:")
        sender_wrap = sender.parent
        sender.decompose()
        sender_email = sender_wrap.contents[0].strip()

        links = message.find_all("a", href=True)
        msg_id_link = links[1]
        msg_id = msg_id_link["href"].split("=")[-1]

        subject_tag = message.find("strong", string="Subject:")
        subject = subject_tag.find_next("td").text.strip()

        size_tag = message.find("strong", string="Size:")
        size = size_tag.find_next("td").text.strip()

        reason_tag = message.find("strong", string="Reason:")
        reason = reason_tag.find_next("td").text.strip()

        received_tag = message.find("strong", string="Received:")
        received_date = received_tag.find_next("td").text.strip()
        received_date_ts = datetime.strptime(received_date, "%a %b %d %H:%M:%S %Y")

        results.append(
            ModerationRequest(
                message_id=int(msg_id),
                sender_email=sender_email,
                subject=subject,
                size_description=size,
                reason=reason,
                received_date=received_date_ts,
            )
        )

    return results


def extract_moderation_post_details(
    message_id: int, raw_html: str
) -> Optional[ModerationDetails]:
    soup = BeautifulSoup(raw_html, "html.parser")

    excerpt_heading = soup.find("strong", string="Message Excerpt:")
    if excerpt_heading is None:
        # This indicates that there is no such post (held for moderation).
        return None

    excerpt_wrap = excerpt_heading.parent.parent.find("textarea")  # type: ignore
    message_contents = excerpt_wrap.contents[0].text  # type: ignore

    headers_heading = soup.find("strong", string="Message Headers:")
    assert headers_heading is not None, "Moderation details HTML structure has changed."
    headers_wrap = headers_heading.parent.parent.find("textarea")  # type: ignore
    headers = headers_wrap.contents[0].text  # type: ignore

    return ModerationDetails(message_id, message_contents, headers)



import enum
from datetime import datetime


class ModerationRequest:
    def __init__(
        self,
        message_id: int,
        sender_email: str,
        subject: str,
        size_description: str,
        reason: str,
        received_date: datetime,
    ) -> None:
        self.message_id = message_id
        self.sender_email = sender_email
        self.subject = subject
        self.size_description = size_description
        self.reason = reason
        self.received_date = received_date

    def __repr__(self) -> str:
        items = ["{}={}".format(k, v) for k, v in self.__dict__.items()]
        return "{}({})".format(self.__class__.__name__, ", ".join(items))


class ModerationDetails:
    def __init__(self, message_id: int, message_contents: str, headers: str) -> None:
        self.message_id = message_id
        self.message_contents = message_contents
        self.headers = headers


class ModerationAction(enum.Enum):
    # NOTE: The values are significant; they correspond to values used by
    # Mailman to represent these actions.

    # Defer is a no-op. We include it for completeness.
    Defer = 0
    Approve = 1
    Reject = 2
    Discard = 3



import requests
from typing import List, Optional

from noire.constants import (
    LOG_IN_URL_TEMPLATE,
    GET_MEMBERS_LIST_URL_TEMPLATE,
    MODERATION_REQUESTS_URL_TEMPLATE,
    MODERATION_DETAILS_URL_TEMPLATE,
    ADD_MEMBERS_URL_TEMPLATE,
    REMOVE_MEMBERS_URL_TEMPLATE,
)
from noire.models.membership import BulkAddResults, BulkRemoveResults
from noire.models.moderation import (
    ModerationRequest,
    ModerationDetails,
    ModerationAction,
)
from noire.parsers.members_list import (
    extract_member_emails,
    extract_add_results,
    extract_remove_results,
)
from noire.parsers.moderation import (
    extract_moderation_requests,
    extract_moderation_post_details,
)


class Noire:
    """
    Provides programmatic access to Mailman 2 via its web user interface.

    This tool is meant to be a modern successor to "mmblanche".
    """

    @classmethod
    def connect(
        cls,
        list_name: str,
        list_password: str,
        mailman_base_url: str,
    ) -> "Noire":
        session = requests.Session()
        log_in_data = {
            "adminpw": list_password,
        }
        log_in_url = LOG_IN_URL_TEMPLATE.format(
            list_name=list_name, mailman_base_url=mailman_base_url
        )
        response = session.post(log_in_url, data=log_in_data)
        if response.status_code == 401:
            raise RuntimeError(f"Incorrect password for list {list_name}")
        elif response.status_code != 200:
            raise RuntimeError(
                f"Error authenticating to list {list_name}: {response.status_code}"
            )

        return cls(list_name, mailman_base_url, list_password, session)

    def __init__(
        self,
        list_name: str,
        mailman_base_url: str,
        list_password: str,
        session: requests.Session,
    ) -> None:
        self._list_name = list_name
        self._mailman_base_url = mailman_base_url
        self._list_password = list_password
        self._session = session

    def get_member_emails(self) -> List[str]:
        """
        Retrieves all emails that are subscribed to the list.
        """
        get_url = GET_MEMBERS_LIST_URL_TEMPLATE.format(
            list_name=self._list_name, mailman_base_url=self._mailman_base_url
        )
        response = self._session.get(get_url)
        if response.status_code != 200:
            raise RuntimeError(
                f"Unexpected error when fetching member emails: {response.status_code}"
            )
        # TODO: Handle pagination for large lists.
        return extract_member_emails(response.content.decode())

    def get_moderation_requests(self) -> List[ModerationRequest]:
        """
        Retrieves moderation requests for the list (i.e., emails sent to the
        list that are held for moderation).
        """
        get_url = MODERATION_REQUESTS_URL_TEMPLATE.format(
            list_name=self._list_name, mailman_base_url=self._mailman_base_url
        )
        response = self._session.get(get_url)
        if response.status_code != 200:
            raise RuntimeError(
                f"Unexpected error when fetching moderation requests: {response.status_code}"
            )
        return extract_moderation_requests(response.content.decode())

    def get_moderation_details(self, message_id: int) -> Optional[ModerationDetails]:
        """
        Retrieves details about a message held for moderation (e.g., the
        message's contents).
        """
        get_url = MODERATION_DETAILS_URL_TEMPLATE.format(
            mailman_base_url=self._mailman_base_url,
            list_name=self._list_name,
            message_id=message_id,
        )
        response = self._session.get(get_url)
        if response.status_code != 200:
            raise RuntimeError(
                f"Unexpected error when fetching moderation details for {message_id}: {response.status_code}"
            )
        return extract_moderation_post_details(message_id, response.content.decode())

    def apply_moderation_action(
        self,
        message_id: int,
        action: ModerationAction,
        rejection_message: Optional[str] = None,
        preserve_message_for_admin: bool = False,
        forward_message_to_list_owner: bool = False,
    ) -> bool:
        """
        Applies a moderation action to a message (defer, reject, approve,
        discard). This method returns true if the moderation action request was
        successfully submitted. The `message_id` should correspond to a message
        ID retrieved from `get_moderation_requests()`.

        If a message is rejected, Mailman will send a rejection reason to the
        sender. You can customize this reason with `rejection_message`. If you
        do not set a rejection message, Mailman will use a default message.
        """

        endpoint = MODERATION_REQUESTS_URL_TEMPLATE.format(
            mailman_base_url=self._mailman_base_url, list_name=self._list_name
        )
        payload = {
            "adminpw": self._list_password,
            message_id: action.value,
        }
        if action == ModerationAction.Reject and rejection_message is not None:
            payload[f"comment-{message_id}"] = rejection_message
        if preserve_message_for_admin:
            payload[f"preserve-{message_id}"] = "on"
        if forward_message_to_list_owner:
            payload[f"forward-{message_id}"] = "on"
        response = self._session.post(endpoint, payload)
        return response.status_code == 200

    def add_members(
        self,
        emails: List[str],
        send_welcome_message: bool = False,
        send_owner_notifications: bool = False,
    ) -> BulkAddResults:
        """
        Subscribes the given emails to the list. Returns the emails that were
        successfully subscribed and the emails that failed to be subscribed.

        Optional arguments:
        - send_welcome_message:      Set to notify the new member that they were
                                     added.
        - send_owner_notifications:  Send an email to the list owner about the new
                                     members.
        """
        endpoint = ADD_MEMBERS_URL_TEMPLATE.format(
            mailman_base_url=self._mailman_base_url, list_name=self._list_name
        )
        payload = {
            "adminpw": self._list_password,
            "subscribees": "\n".join(emails),
            "subscribe_or_invite": 0,  # 0 indicates subscribe.
            "send_welcome_msg_to_this_batch": 1 if send_welcome_message else 0,
            "send_notifications_to_list_owner": 1 if send_owner_notifications else 0,
        }
        response = self._session.post(endpoint, payload)
        return extract_add_results(response.content.decode())

    def remove_members(
        self,
        emails: List[str],
        send_unsubscribe_message: bool = False,
        send_owner_notifications: bool = False,
    ) -> BulkRemoveResults:
        """
        Unsubscribes the given emails from the list. Returns the emails that were
        successfully removed.

        Optional arguments:
        - send_unsubscribe_message: Set to notify the member that they were removed.
        - send_owner_notifications: Send an email to the list owner about the removed
                                    members.
        """
        endpoint = REMOVE_MEMBERS_URL_TEMPLATE.format(
            mailman_base_url=self._mailman_base_url, list_name=self._list_name
        )
        payload = {
            "adminpw": self._list_password,
            "unsubscribees": "\n".join(emails),
            "send_unsub_ack_to_this_batch": 1 if send_unsubscribe_message else 0,
            "send_unsub_notifications_to_list_owner": 1
            if send_owner_notifications
            else 0,
        }
        response = self._session.post(endpoint, payload)
        return extract_remove_results(response.content.decode())



from bs4 import BeautifulSoup
from typing import List

from noire.models.membership import BulkAddResults, MemberError, BulkRemoveResults


def extract_member_emails(raw_html: str) -> List[str]:
    """
    Extracts the member emails that appear on `/mailman/admin/<list name>/members`
    from the raw HTML.
    """
    table_index = 4
    member_emails: List[str] = []

    soup = BeautifulSoup(raw_html, "html.parser")

    # Find all the tables on the page
    tables = soup.find_all("table")

    if table_index >= len(tables):
        return member_emails

    table = tables[table_index]
    rows = table.find_all("tr")

    # Skip the first row since it contains the table headers.
    for row in rows[1:]:
        # Find the first 'a' element to get the email address.
        email_element = row.find("a")
        if email_element:
            email = email_element.text.strip()
            member_emails.append(email)

    return member_emails


def extract_add_results(
    raw_html: str,
) -> BulkAddResults:
    soup = BeautifulSoup(raw_html, "html.parser")

    # Extract emails under "Successfully subscribed"
    success_subscribing = soup.find("h5", string="Successfully subscribed:")
    if success_subscribing is not None:
        success_emails = _extract_from_malformed_li(
            success_subscribing.find_next("ul").decode_contents()  # type: ignore
        )
    else:
        success_emails = []

    # Extract emails under "Error subscribing"
    error_subscribing = soup.find("h5", string="Error subscribing:")
    error_emails = []

    if error_subscribing is not None:
        cleaned_errors = _extract_from_malformed_li(
            error_subscribing.find_next("ul").decode_contents()  # type: ignore
        )
        for item in cleaned_errors:
            parts = item.split(" -- ")
            if len(parts) > 1:
                error_emails.append(MemberError(parts[0], parts[1]))
            else:
                error_emails.append(MemberError(parts[0], None))

    return BulkAddResults(success_emails, error_emails)


def extract_remove_results(raw_html: str) -> BulkRemoveResults:
    soup = BeautifulSoup(raw_html, "html.parser")

    # Extract emails under "Successfully unsubscribed"
    success_removed = soup.find("h5", string="Successfully Unsubscribed:")
    if success_removed is not None:
        removed = _extract_from_malformed_li(
            success_removed.find_next("ul").decode_contents()  # type: ignore
        )
    else:
        removed = []

    return BulkRemoveResults(removed)


def _extract_from_malformed_li(raw_content: str) -> List[str]:
    # Mailman's lists are malformed (they are missing </li> tags). This function
    # works around the malformed HTML and extracts the intended list items.
    remove_trailing_li = raw_content.replace("</li>", "")
    items = []
    for piece in remove_trailing_li.split("<li>"):
        clean = piece.strip()
        if len(clean) == 0:
            continue
        items.append(clean)
    return items



from typing import List, Optional


class MemberError:
    def __init__(self, email: str, error_reason: Optional[str]) -> None:
        self.email = email
        self.error_reason = error_reason

    def __repr__(self) -> str:
        return "".join(["(", self.email, ", reason=", str(self.error_reason), ")"])


class BulkAddResults:
    def __init__(self, added: List[str], errors: List[MemberError]) -> None:
        self.added = added
        self.errors = errors

    def __repr__(self) -> str:
        added = "\n    ".join(self.added)
        errors = "\n    ".join(map(repr, self.errors))
        return "\n".join(
            ["BulkAddResults", "  Added:", "    " + added, "  Errors:", "    " + errors]
        )


class BulkRemoveResults:
    def __init__(self, removed: List[str]) -> None:
        self.removed = removed

    def __repr__(self) -> str:
        removed = "\n    ".join(self.removed)
        return "\n".join(["BulkRemoveResults", "  Removed:", "    " + removed])



import codecs
import os
import re

from setuptools import setup, find_packages

# Acknowledgement: This setup.py was adapted from Hynek Schlawack's Python
#                  Packaging Guide
# https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty

###################################################################

NAME = "noire"
PACKAGES = find_packages(where="src")
META_PATH = os.path.join("src", "noire", "__init__.py")
README_PATH = "README.md"
PYTHON_REQUIRES = ">=3.10"

PACKAGE_DATA = {}
PACKAGE_DIR = {"": "src"}

ENTRY_POINTS = {}

INSTALL_REQUIRES = [
    "beautifulsoup4>=4.12.2",
    "requests",
]

DEV_REQUIRES = [
    "black",
    "mypy",
    "pep517",
    "pylint",
    "types-beautifulsoup4",
    "types-requests",
]

KEYWORDS = []

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3 :: Only",
]

###################################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and return the contents of the
    resulting file. Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


if __name__ == "__main__":
    setup(
        name=NAME,
        description=find_meta("description"),
        version=find_meta("version"),
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        url=find_meta("url"),
        long_description=read(README_PATH),
        long_description_content_type="text/markdown",
        packages=PACKAGES,
        package_data=PACKAGE_DATA,
        package_dir=PACKAGE_DIR,
        python_requires=PYTHON_REQUIRES,
        install_requires=INSTALL_REQUIRES,
        extras_require={
            "dev": DEV_REQUIRES,
        },
        entry_points=ENTRY_POINTS,
        classifiers=CLASSIFIERS,
        keywords=KEYWORDS,
    )



LOG_IN_URL_TEMPLATE = "{mailman_base_url}/admin/{list_name}"

GET_MEMBERS_LIST_URL_TEMPLATE = "{mailman_base_url}/admin/{list_name}/members"
MODERATION_REQUESTS_URL_TEMPLATE = "{mailman_base_url}/admindb/{list_name}"
MODERATION_DETAILS_URL_TEMPLATE = (
    "{mailman_base_url}/admindb/{list_name}?msgid={message_id}"
)
ADD_MEMBERS_URL_TEMPLATE = "{mailman_base_url}/admin/{list_name}/members/add"
REMOVE_MEMBERS_URL_TEMPLATE = "{mailman_base_url}/admin/{list_name}/members/remove"



from .noire import Noire

__version__ = "0.1.0"

__author__ = "Geoffrey Yu"
__email__ = "geoffxy@gmail.com"

__description__ = "Programmatic access to Mailman via its publicly-accessible web interface."
__url__ = "https://github.com/geoffxy/noire/"
